#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import zipfile
import shutil
import pathlib
import uuid
import octobot_commons.constants as constants

# avoid cyclic import
from octobot_commons.profiles.profile import Profile


def export_profile(profile, export_path: str) -> str:
    """
    Exports the given profile into export_path, appends ".zip" as a file extension
    :param profile: profile to export
    :param export_path: export path ending with filename
    :return: the exported profile path including file extension
    """
    export_path_with_ext = f"{export_path}.{constants.PROFILE_EXPORT_FORMAT}"
    # remove any existing file to prevent any side effect
    if os.path.isfile(export_path_with_ext):
        os.remove(export_path_with_ext)
    shutil.make_archive(export_path, constants.PROFILE_EXPORT_FORMAT, profile.path)
    return export_path_with_ext


def import_profile(
    import_path: str, name: str = None, bot_install_path: str = "."
) -> None:
    """
    Imports the given profile export archive into the user's profile directory with the "imported_" prefix
    :param import_path: path to the profile zipped archive
    :param name: name of the profile folder
    :param bot_install_path: path to the octobot installation
    :return: None
    """
    profile_name = name or (
        f"{constants.IMPORTED_PROFILE_PREFIX}_{os.path.split(import_path)[-1]}"
    )
    profile_name = profile_name.split(f".{constants.PROFILE_EXPORT_FORMAT}")[0]
    target_import_path = os.path.join(
        bot_install_path, constants.USER_PROFILES_FOLDER, profile_name
    )
    target_import_path = _get_unique_profile_folder(target_import_path)
    if zipfile.is_zipfile(import_path):
        with zipfile.ZipFile(import_path) as zipped:
            zipped.extractall(target_import_path)
    else:
        shutil.copytree(import_path, target_import_path)
    _ensure_unique_profile_id(target_import_path)


def _get_unique_profile_folder(target_import_path):
    iteration = 1
    candidate = target_import_path
    while os.path.exists(candidate) and iteration < 100:
        iteration += 1
        candidate = f"{target_import_path}_{iteration}"
    return candidate


def _ensure_unique_profile_id(profile_path):
    ids = Profile.get_all_profiles_ids(
        pathlib.Path(profile_path).parent, ignore=profile_path
    )
    profile = Profile(profile_path)
    profile.read_config()
    iteration = 1
    while profile.profile_id in ids and iteration < 100:
        profile.profile_id = str(uuid.uuid4())
        iteration += 1
    profile.save()



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants


DELETE_ELEMENT_VALUE = ""


def filter_to_update_data(to_update_data, in_backtesting):
    """
    Filter data to update
    :param to_update_data: the data to be updated
    :param in_backtesting: if backtesting is enabled
    :return: the updated data
    """
    if in_backtesting:
        for key in set(to_update_data.keys()):
            # remove changes to currency config when in backtesting
            if commons_constants.CONFIG_CRYPTO_CURRENCIES in key:
                to_update_data.pop(key)


def parse_and_update(key, new_data, config_separator):
    """
    Parse and update key
    :param key: the key to update
    :param new_data: the new data
    :param config_separator: the config separator
    :return: the key updated
    """
    parsed_data_array = key.split(config_separator)
    new_config = {}
    current_dict = new_config

    for i, _ in enumerate(parsed_data_array):
        if i > 0:
            if i == len(parsed_data_array) - 1:
                current_dict[parsed_data_array[i]] = new_data
            else:
                current_dict[parsed_data_array[i]] = {}
        else:
            new_config[parsed_data_array[i]] = {}

        current_dict = current_dict[parsed_data_array[i]]

    return new_config


def merge_dictionaries_by_appending_keys(dict_dest, dict_src) -> dict:
    """
    Merge dictionnaries by appending keys
    :param dict_dest: the destination dictionnary
    :param dict_src: the source dictionnary
    :return: the merged dictionnary
    """
    for key in dict_src:
        src_val = dict_src[key]
        if key in dict_dest:
            dest_val = dict_dest[key]
            if isinstance(dest_val, dict) and isinstance(src_val, dict):
                dict_dest[key] = merge_dictionaries_by_appending_keys(dest_val, src_val)
            elif dest_val == src_val:
                pass  # same leaf value
            elif _are_of_compatible_type(dest_val, src_val):
                # simple type: update value
                dict_dest[key] = src_val
            elif isinstance(dest_val, list) and isinstance(src_val, list):
                dict_dest[key] = src_val
            else:
                logging.get_logger("ConfigOperations").error(
                    f"Conflict when merging dict with key : {key}"
                )
        else:
            dict_dest[key] = src_val

    return dict_dest


def clear_dictionaries_by_keys(dict_dest, dict_src):
    """
    Clear dictionnaries by keys
    :param dict_dest: the destination dictionnary
    :param dict_src: the source dictionnary
    :return: the cleaned dictionnary
    """
    for key in dict_src:
        src_val = dict_src[key]
        if key in dict_dest:
            dest_val = dict_dest[key]
            if src_val == DELETE_ELEMENT_VALUE:
                dict_dest.pop(key)
            elif isinstance(dest_val, dict) and isinstance(src_val, dict):
                dict_dest[key] = clear_dictionaries_by_keys(dest_val, src_val)
            else:
                logging.get_logger("ConfigOperations").error(
                    f"Conflict when deleting dict element with key : {key}"
                )

    return dict_dest


def _are_of_compatible_type(val1, val2) -> bool:
    """
    Check if types are compatibles
    :param val1: the first value
    :param val2: the second value
    :return: True if types are compatible
    """
    return (
        isinstance(val1, val2.__class__)
        or (isinstance(val1, (float, int)) and isinstance(val2, (float, int)))
    ) and isinstance(val1, (bool, str, float, int))



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.logging import logging_util
from octobot_commons.logging.logging_util import (
    BotLogger,
    set_global_logger_level,
    get_global_logger_level,
    get_logger,
    set_logging_level,
    get_backtesting_errors_count,
    reset_backtesting_errors,
    set_error_publication_enabled,
    BACKTESTING_NEW_ERRORS_COUNT,
    LOG_DATABASE,
    LOG_NEW_ERRORS_COUNT,
    logs_database,
    error_notifier_callbacks,
    LOGS_MAX_COUNT,
    add_log,
    get_errors_count,
    reset_errors_count,
    register_error_notifier,
)

__all__ = [
    "BotLogger",
    "set_global_logger_level",
    "get_global_logger_level",
    "get_logger",
    "set_logging_level",
    "get_backtesting_errors_count",
    "reset_backtesting_errors",
    "set_error_publication_enabled",
    "BACKTESTING_NEW_ERRORS_COUNT",
    "LOG_DATABASE",
    "LOG_NEW_ERRORS_COUNT",
    "logs_database",
    "error_notifier_callbacks",
    "LOGS_MAX_COUNT",
    "add_log",
    "get_errors_count",
    "reset_errors_count",
    "register_error_notifier",
]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


async def download_stream_file(
    output_file,
    file_url,
    aiohttp_session,
    data_chunk_size=5 * 2 ** 20,
    is_aiofiles_output_file=False,
):
    """
    Download a big file with an aiohttp session
    :param output_file: the output file
    :param file_url: the file to be downloaded url
    :param aiohttp_session: the aiohttp session
    :param data_chunk_size: default value is 5*2**20 (5MB)
    :param is_aiofiles_output_file: When True, output_file.write will be awaited (when it's an aiofiles instance)
    """
    async with aiohttp_session.get(file_url) as resp:
        if resp.status != 200:
            raise RuntimeError(
                f"Failed to download file at url : {file_url} (status: {resp.status})"
            )
        while True:
            chunk = await resp.content.read(data_chunk_size)
            if not chunk:
                # resp.content.read returns an empty chunk when completed
                break
            if is_aiofiles_output_file:
                await output_file.write(chunk)
            else:
                output_file.write(chunk)



# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cdef class EventTreeNode:
    cdef public object node_value
    cdef public object node_value_time
    cdef public object node_type
    cdef public dict children

cdef class NodeExistsError(Exception):
    pass

cdef class EventTree:
    cdef public EventTreeNode root

    cpdef void set_node(self, object value, object node_type, EventTreeNode node, double timestamp=*)
    cpdef void set_node_at_path(self, object value, object node_type, list path, double timestamp=*)
    cpdef EventTreeNode get_node(self, list path, EventTreeNode starting_node=*)
    cpdef EventTreeNode get_or_create_node(self, list path, EventTreeNode starting_node=*)

    cdef EventTreeNode _get_node(self, list path, EventTreeNode starting_node=*)
    cdef EventTreeNode _create_node_path(self, list path, EventTreeNode starting_node=*)
    cdef void _set_node(self, EventTreeNode node, object value=*, object node_type=*, double timestamp=*)



class Singleton:
    """
    From https://stackoverflow.com/questions/51245056/singleton-is-not-working-in-cython
    """

    _instances = {}

    @classmethod
    def instance(cls, *args, **kwargs):
        """
        Create the instance if not already created
        Return the class instance
        :param args: the constructor arguments
        :param kwargs: the constructor optional arguments
        :return: the class only instance
        """
        if cls not in cls._instances:
            cls._instances[cls] = cls(*args, **kwargs)
        return cls._instances[cls]



# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef class AsyncJob:
    cdef object logger                      # logging.Logger
    cdef object callback                    # Callable
    cdef object job_task                    # Asyncio.Task
    cdef object job_periodic_task           # Asyncio.Task
    cdef public object idle_task_event      # Asyncio.Event

    cdef bint is_started
    cdef bint should_stop
    cdef bint is_periodic
    cdef bint enable_multiple_runs
    cdef int simultaneous_calls
    cdef int successive_failures
    cdef int max_successive_failures

    cdef double last_execution_time
    cdef double execution_interval_delay
    cdef double min_execution_delay

    cdef list job_dependencies

    cpdef void add_job_dependency(self, AsyncJob job)
    cpdef bint is_job_idle(self)
    cpdef void clear(self)
    cpdef void stop(self)

    cdef void _handle_run_exception(self, object exception, bint error_on_single_failure)
    cdef bint _should_run_job(self, bint force=*, ignore_dependencies=*)
    cdef bint _has_enough_time_elapsed(self)
    cdef bint _are_job_dependencies_idle(self)
